"""Classes for the CloudWanderer DynamoDB Storage Connector."""
import logging
import os
import pathlib
import boto3
import json
from datetime import datetime
from random import randrange
from decimal import Decimal
from .base_connector import BaseConnector
from boto3.dynamodb.conditions import Key
from ..cloud_wanderer import ResourceDict
from ..aws_urn import AwsUrn


def gen_resource_type_index(service, resource_type):
    """Generate a primary key for the resource type index."""
    return f"{service}#{resource_type}"


def gen_shard(key, shard_id=None):
    """Append a shard designation to the end of a supplied key."""
    shard_id = shard_id if shard_id is not None else randrange(10)
    return f"{key}#shard{shard_id}"


def primary_key_from_urn(urn):
    """Create a DynamoDB Primary Key from a resource's URN."""
    return f"resource#{urn}"


def urn_from_primary_key(pk):
    """Create an AwsUrn from a resource's primary key."""
    return AwsUrn.from_string(pk.split('#')[1])


def dynamodb_items_to_resources(items):
    """Convert a resource item retrieved from DynamoDB to a ResourceDict."""
    for item in items:
        yield ResourceDict(
            urn=urn_from_primary_key(item['_id']),
            resource=item
        )


def json_default(item):
    """JSON object type converter that handles datetime objects."""
    if isinstance(item, datetime):
        return item.isoformat()


def standardise_data_types(resource):
    """Return a dictionary normalised to datatypes acceptable for DynamoDB."""
    result = json.loads(json.dumps(resource, default=json_default), parse_float=Decimal)
    return result


class DynamoDbConnector(BaseConnector):
    """CloudWanderer Storage Connector for DynamoDB.

    Arguments:
        table_name (str): The name of the table to store resources in.
        endpoint_url (str): optional override endpoint url for DynamoDB.
    """

    def __init__(self, table_name='cloud_wanderer', endpoint_url=None):
        """Initialise the DynamoDbConnector."""
        self.endpoint_url = endpoint_url
        self.table_name = table_name
        self.dynamodb = boto3.resource('dynamodb', endpoint_url=endpoint_url)
        self.dynamodb_table = self.dynamodb.Table(table_name)

    def init(self):
        """Create the DynamoDB Database."""
        table_creator = DynamoDbTableCreator(
            boto3_dynamodb_resource=self.dynamodb,
            table_name=self.table_name
        )
        table_creator.create_table()

    def write(self, urn, resource):
        """Write the specified resource to DynamoDB.

        Arguments:
            urn (cloudwanderer.AwsUrn): The URN of the resource.
            resource: The boto3 Resource object representing the resource.
        """
        logging.debug(f"Writing: {urn} to {self.table_name}")
        item = {
            **{
                '_id': primary_key_from_urn(urn),
                '_urn': str(urn),
                '_resource_type': f"{gen_resource_type_index(urn.service, urn.resource_type)}",
                '_resource_type_index': f"{gen_shard(gen_resource_type_index(urn.service, urn.resource_type))}",
                '_account_id': f"{urn.account_id}",
                '_account_id_index': f"{gen_shard(urn.account_id)}"
            },
            **standardise_data_types(resource.meta.data or {})
        }
        self.dynamodb_table.put_item(
            Item=item
        )

    def read_resource(self, urn):
        """Return the resource with the specified :class:`cloudwanderer.AwsUrn`.

        Arguments:
            urn (cloudwanderer.AwsUrn): The AWS URN of the resource to return
        """
        result = self.dynamodb_table.query(
            KeyConditionExpression=Key('_id').eq(primary_key_from_urn(urn))
        )
        yield from dynamodb_items_to_resources(result['Items'])

    def read_resource_of_type(self, service, resource_type):
        """Return all resources of type.

        Args:
            service (str): Service name (e.g. ec2)
            resource_type (str): Resource Type (e.g. instance)
        """
        for shard_id in range(0, 9):
            key = gen_shard(gen_resource_type_index(service, resource_type), shard_id)
            logging.debug("Fetching shard %s", key)
            result = self.dynamodb_table.query(
                IndexName='resource_type',
                Select='ALL_PROJECTED_ATTRIBUTES',
                KeyConditionExpression=Key('_resource_type_index').eq(key)
            )
            yield from dynamodb_items_to_resources(result['Items'])

    def read_all_resources_in_account(self, account_id):
        """Return all resources in account.

        Args:
            account_id (str): AWS Account ID
        """
        for shard_id in range(0, 9):
            key = gen_shard(account_id, shard_id)
            logging.debug("Fetching shard %s", key)
            result = self.dynamodb_table.query(
                IndexName='account_id',
                Select='ALL_PROJECTED_ATTRIBUTES',
                KeyConditionExpression=Key('_account_id_index').eq(key)
            )
            yield from dynamodb_items_to_resources(result['Items'])

    def read_resource_of_type_in_account(self, service, resource_type, account_id):
        """Return all resources of the specified type in the specified AWS account.

        Args:
            service (str): Service name, e.g. ``ec2``
            resource_type (str): Resouce type, e.g. ``instance``
            account_id (str): AWS Account ID
        """
        for shard_id in range(0, 9):
            key = gen_shard(account_id, shard_id)
            logging.debug("Fetching shard %s", key)
            result = self.dynamodb_table.query(
                IndexName='account_id',
                Select='ALL_PROJECTED_ATTRIBUTES',
                KeyConditionExpression=(
                    Key('_account_id_index').eq(key)
                    & Key('_resource_type').eq(gen_resource_type_index(service, resource_type))
                )
            )
            yield from dynamodb_items_to_resources(result['Items'])

    def read_all(self):
        """Return all DynamoDB table records (not just resources)."""
        return dynamodb_items_to_resources(self.dynamodb_table.scan()['Items'])


class DynamoDbTableCreator():
    """DynamoDB Table Creator class.

    Arguments:
        boto3_dynamodb_resource:
            The dynamodb resource object from boto3.
        table_name (str):
            The name of the table to create.

    """

    schema_file = os.path.join(
        pathlib.Path(__file__).parent.absolute(),
        'dynamodb_schema.json'
    )

    def __init__(self, boto3_dynamodb_resource, table_name):
        """Initialise the DynamoDB Table Creator."""
        self.dynamodb = boto3_dynamodb_resource
        self.dynamodb_table = self.dynamodb.Table(table_name)
        self.table_name = table_name
        self._schema = None

    def create_table(self):
        """Create the DynamoDB table."""
        try:
            self.dynamodb.create_table(**{
                **self.schema['table'],
                **{'TableName': self.table_name}
            })
        except self.dynamodb_table.meta.client.exceptions.ResourceInUseException:
            logging.warning(
                'Table %s already exists, skipping creation.',
                self.table_name)

    @property
    def schema(self):
        """Return the DynamoDB Schema."""
        if not self._schema:
            with open(self.schema_file) as schema_file:
                self._schema = json.load(schema_file)
        return self._schema



"""Main cloudwanderer module."""
import logging
from botocore import xform_name
from botocore.exceptions import ClientError
from boto3.exceptions import ResourceNotExistsError
import boto3
from .custom_resource_definitions import CustomResourceDefinitions
from .aws_urn import AwsUrn
GLOBAL_SERVICE_REGIONAL_RESOURCE = [
    {
        'resource_name': 's3_bucket'
    }
]


class CloudWandererBoto3Interface():
    """Class of methods which expect boto3 resources and services rather than resource names and service names."""

    def __init__(self):
        """Class of methods which expect boto3 resources and services rather than resource names and service names."""
        self.custom_resource_definitions = CustomResourceDefinitions().load_custom_resource_definitions()

    def _get_available_services(self):
        return boto3.Session().get_available_resources()

    def get_all_resource_services(self):
        """Return all the boto3 service Resource objects that are available, both built-in and custom."""
        for service_name in self._get_available_services():
            yield self.get_boto3_resource_service(service_name)
        for service_name in self.custom_resource_definitions:
            yield self.get_custom_resource_service(service_name)

    def get_boto3_resource_service(self, service_name):
        """Return the boto3 service Resource object matching this service_name."""
        try:
            return boto3.resource(service_name)
        except ResourceNotExistsError:
            return None

    def get_custom_resource_service(self, service_name):
        """Get the custom resource definition matching this service name."""
        return self.custom_resource_definitions.get(service_name)

    def get_resource_service_by_name(self, service_name):
        """Return all services matching name, boto3 or custom."""
        boto3_resource_service = self.get_boto3_resource_service(service_name)
        if boto3_resource_service:
            yield boto3_resource_service
        custom_resource_service = self.get_custom_resource_service(service_name)
        if custom_resource_service:
            yield custom_resource_service

    def get_resource_collections(self, boto3_service):
        """Return all resource types in this service."""
        return boto3_service.meta.resource_model.collections

    def get_resource_from_collection(self, boto3_service, boto3_resource_collection):
        """Return all resources of this resource type (collection) from this service."""
        logging.info(f'--> Fetching {boto3_service.meta.service_name} {boto3_resource_collection.name}')
        try:
            for resource in getattr(boto3_service, boto3_resource_collection.name).all():
                yield resource
        except ClientError as ex:
            if ex.response['Error']['Code'] == 'InvalidAction':
                logging.warning(ex.response['Error']['Message'])
                return
            raise ex


class CloudWanderer():
    """CloudWanderer.

    Args:
        storage_connector: A CloudWanderer storage connector object.
    """

    def __init__(self, storage_connector):
        """Initialise CloudWanderer."""
        self.storage_connector = storage_connector
        self.boto3_interface = CloudWandererBoto3Interface()
        self._account_id = None
        self._client_region = None

    def write_all_resources(self, exclude_resources=None):
        """Write all AWS resources in this account from all services to storage."""
        exclude_resources = exclude_resources or []
        for boto3_service in self.boto3_interface.get_all_resource_services():
            for boto3_resource_collection in self.boto3_interface.get_resource_collections(boto3_service):
                if boto3_resource_collection.name in exclude_resources:
                    logging.info('Skipping %s as per exclude_resources', boto3_resource_collection.name)
                    continue
                resources = self.boto3_interface.get_resource_from_collection(
                    boto3_service,
                    boto3_resource_collection
                )
                for boto3_resource in resources:
                    self.storage_connector.write(self._get_resource_urn(boto3_resource), boto3_resource)

    def write_resources(self, service_name, exclude_resources=None):
        """Write all AWS resources in this account in this service to storage.

        Arguments:
            service_name (str): The name of the service to write resources for (e.g. ``'ec2'``)
            exclude_resources (list): A list of resources to exclude (e.g. `['instances']`)
        """
        for boto3_service in self.boto3_interface.get_resource_service_by_name(service_name):
            for boto3_resource_collection in self.boto3_interface.get_resource_collections(boto3_service):
                if boto3_resource_collection.name in exclude_resources:
                    logging.info('Skipping %s as per exclude_resources', boto3_resource_collection.name)
                    continue
                resources = self.boto3_interface.get_resource_from_collection(boto3_service, boto3_resource_collection)
                for boto3_resource in resources:
                    self.storage_connector.write(self._get_resource_urn(boto3_resource), boto3_resource)

    def get_resources(self, boto3_service, exclude_resources):
        """Return all resources for this service from the AWS API.

        Arguments:
            boto3_service: The ``boto3.resource()`` service to get the resources of.
            exclude_resources (list): A list of resources to exclude (e.g. ``['instances']``)
        """
        for collection in self.get_resource_collections(boto3_service):
            logging.info(f'--> Fetching {boto3_service.meta.service_name} {collection.name}')

            if collection.name in ['images', 'snapshots']:
                continue
            try:
                for resource in getattr(boto3_service, collection.name).all():
                    yield resource
            except ClientError as ex:
                if ex.response['Error']['Code'] == 'InvalidAction':
                    logging.warning(ex.response['Error']['Message'])
                    continue
                raise ex

    def read_resource_of_type(self, service, resource_type):
        """Return all resources of type.

        Args:
            service (str): Service name (e.g. ec2)
            resource_type (str): Resource Type (e.g. instance)
        """
        return self.storage_connector.read_resource_of_type(service, resource_type)

    def read_resource(self, urn):
        """Return a specific resource by its urn from storage."""
        try:
            return next(self.storage_connector.read_resource(urn))
        except StopIteration:
            return None

    def read_all_resources_in_account(self, account_id):
        """Return all resources in the provided AWS Account from storage."""
        return self.storage_connector.read_all_resources_in_account(account_id)

    def read_resource_of_type_in_account(self, service, resource_type, account_id):
        """Return all resources of this type in the provided AWS Account from storage."""
        return self.storage_connector.read_resource_of_type_in_account(service, resource_type, account_id)

    @property
    def account_id(self):
        """Return the AWS Account ID our boto3 client is authenticated against."""
        if self._account_id is None:
            sts = boto3.client('sts')
            self._account_id = sts.get_caller_identity()['Account']
        return self._account_id

    @property
    def client_region(self):
        """Return the region our boto3 client is authenticated against."""
        if self._client_region is None:
            self._client_region = boto3.session.Session().region_name
        return self._client_region

    def _get_resource_urn(self, resource):
        id_member_name = resource.meta.resource_model.identifiers[0].name
        resource_id = getattr(resource, id_member_name)
        if resource_id.startswith('arn:'):
            resource_id = ''.join(resource_id.split(':')[5:])
        return AwsUrn(
            account_id=self.account_id,
            region=self.client_region,
            service=resource.meta.service_name,
            resource_type=xform_name(resource.meta.resource_model.name),
            resource_id=resource_id
        )


class ResourceDict(dict):
    """A dictionary representation of a resource that prevents any storage metadata polluting the resource dictionary.

    Use ``dict(my_resource_dict)`` to convert this object into a dictionary that
    contains *only* the resource's metadata.

    Attributes:
        urn (cloudwanderer.AwsUrn): The AWS URN of the resource.
        metadata (dict): The original storage representation of the resource as it was passed in.
    """

    def __init__(self, urn, resource):
        """Initialise the resource dict."""
        self.urn = urn
        self.metadata = resource
        super().__init__(**self._clean_resource)

    @property
    def _clean_resource(self):
        return {
            key: value
            for key, value in self.metadata.items()
            if not key.startswith('_')
        }

    def __repr__(self):
        """Return a code representation of this resource."""
        return f"{self.__class__.__name__}(urn={repr(self.urn)}, resource={self.metadata})"

    def __str__(self):
        """Return the string representation of this Resource."""
        return repr(self)



"""Classes for handling custom boto3 Resources."""
import os
import json
import pathlib
import botocore
from boto3.resources.factory import ResourceFactory
from boto3.utils import ServiceContext
import boto3


class CustomResourceFactory():
    """Factory class for generating custom boto3 Resource objects."""

    def __init__(self):
        """Initialise the ResourceFactory."""
        self.emitter = boto3.Session().events
        self.factory = ResourceFactory(self.emitter)
        self.botocore_session = botocore.session.get_session()

    def load(self, service_name, resource_definitions=None, service_definition=None):
        """Load the specified resource definition dictionaries into a Resource object."""
        service_context = ServiceContext(
            service_name=service_name,
            resource_json_definitions=resource_definitions,
            service_model=self._get_service_model(service_name),
            service_waiter_model=None
        )

        return self.factory.load_from_definition(
            resource_name=service_name,
            single_resource_json_definition=service_definition,
            service_context=service_context
        )

    def _get_shape(self, service, shape_name):
        service_model = self._get_service_model()
        return service_model.shape_for(shape_name)

    def _get_service_model(self, service):
        client = self.botocore_session.create_client(service)
        return client.meta.service_model


class CustomResourceDefinitions():
    """Custom Resource Definitions.

    Allows us to specify resource definitions where they are not supplied by boto3.
    """

    def __init__(self, definition_path='resource_definitions'):
        """Initialise the CustomResourceDefinition."""
        self.service_definitions_path = os.path.join(
            pathlib.Path(__file__).parent.absolute(),
            definition_path
        )
        self.factory = CustomResourceFactory()

    def load_custom_resource_definitions(self):
        """Return our custom resource definitions."""
        services = {}
        for service_name in self._list_service_definitions():
            service_definition = self._load_service_definition(service_name)
            services[service_name] = self.factory.load(
                service_name=service_name,
                service_definition=service_definition['service'],
                resource_definitions=service_definition['resources'],
            )()
        return services

    def _load_service_definition(self, service_name):
        with open(os.path.join(self.service_definitions_path, f"{service_name}.json")) as definition_path:
            return json.load(definition_path)

    def _list_service_definitions(self):
        return [
            file_name.replace('.json', '')
            for file_name in os.listdir(self.service_definitions_path)
            if os.path.isfile(os.path.join(self.service_definitions_path, file_name))
        ]



"""Module containing abstract classes for CloudWanderer storage connectors."""
from abc import ABC, abstractmethod


class BaseConnector(ABC):
    """Abstract class for specification of the CloudWanderer storage connector interface."""

    @abstractmethod
    def write(self, urn, resource):
        """Persist a single resource to storage."""

    @abstractmethod
    def read_all(self):
        """Return all records from storage."""

    @abstractmethod
    def read_resource(self, urn):
        """Return a resource matching the supplied urn from storage."""

    @abstractmethod
    def read_resource_of_type(self, service, resource_type):
        """Return all resources of this type from storage."""

    @abstractmethod
    def read_all_resources_in_account(self, account_id):
        """Return all resources from this AWS account."""

    @abstractmethod
    def read_resource_of_type_in_account(self, service, resource_type, account_id):
        """Return all resources of this type from this AWS account."""



"""I wandered lonely through the cloud."""
from . import storage_connectors
from .cloud_wanderer import CloudWanderer
from .aws_urn import AwsUrn
from .custom_resource_definitions import CustomResourceDefinitions
__all__ = [
    'storage_connectors',
    'CloudWanderer',
    'AwsUrn',
    'CustomResourceDefinitions'
]



#!/usr/bin/env python
from os import path
from setuptools import setup

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    version='0.2.0',
    python_requires='>=3.6.0',
    name='cloudwanderer',
    packages=['cloudwanderer'],
    description='A Python package which wanders across your AWS account and records your resources in DynamoDB',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Sam Martin',
    author_email='samjackmartin+cloudwanderer@gmail.com',
    url='https://github.com/CloudWanderer-io/CloudWanderer',
    install_requires=['boto3'],
    package_data={
        "": [
            "service_definitions/*.json",
            "storage_connectors/*.json"
        ],
    }

)



"""A dataclass for building and querying AWS URNs."""


class AwsUrn():
    """A dataclass for building and querying AWS URNs.

    Args:
        account_id (str): AWS Account ID (e.g. ``111111111111``).
        region (str): AWS region (e.g. ``eu-west-1``).
        service (str): AWS Service (e.g. ``ec2``).
        resource_type (str): AWS Resource Type (e.g. ``instance``)
        resource_id (str): AWS Resource Id (e.g. ``i-11111111``)
    """

    def __init__(self, account_id, region, service, resource_type, resource_id):
        """Initialise an AWS Urn."""
        self.account_id = account_id
        self.region = region
        self.service = service
        self.resource_type = resource_type
        self.resource_id = resource_id

    @classmethod
    def from_string(cls, urn_string):
        """Create an AwsUrn Object from an AwsUrn string."""
        parts = urn_string.split(':')
        return cls(
            account_id=parts[2],
            region=parts[3],
            service=parts[4],
            resource_type=parts[5],
            resource_id=parts[6]
        )

    def __eq__(self, other):
        """Allow comparison of one AwsUrn to another."""
        return str(self) == str(other)

    def __repr__(self):
        """Return a class representation of the AwsUrn."""
        return str(
            f"{self.__class__.__name__}("
            f"account_id='{self.account_id}', '"
            f"region='{self.region}', '"
            f"service='{self.service}', '"
            f"resource_type='{self.resource_type}', '"
            f"resource_id='{self.resource_id}')"
        )

    def __str__(self):
        """Return a string representation of the AwsUrn."""
        return str(
            f"urn:aws:{self.account_id}:{self.region}:{self.service}:{self.resource_type}:{self.resource_id}"
        )
